/*
 * index.c - Parsing of index (and zone configuration) files.
 */
#include "fuzzy.h"
#include "index.h"
#include "dio.h"
#include "dfp.h"
#include "dbf.h"
#include "mlog.h"

void *diku_index_kludge(const char *filename);

/*
 * MODULE DESCRIPTION
 *
 * Nearly every data file loader will need extra parameters to accomplish its goal
 * of providing data in request for a zone.  The SQL will need host, port, username,
 * password, database, and query just to get information back.  An FTP module might
 * need FTP IP, username, password, path.  The mmap and fread modules will only need
 * filename.  This will also make loading every zone listed in an index (where each
 * zone may have different methods) easy.
 *
 * This is quite similar to my previous implementation (but not based upon it) 
 * because I believe the original design idea was sound.  Reimplementations suck.
 *
 * -gg 6/10/99
 */

/*
 * Objective: Cycle through lines in file one by one and
 * invoke the single line parser.  Not too complicated.
 */
void index_load(World *world, const char *file)
{
  char *input_line = NULL;
  size_t string_length = 0;
  ssize_t retlength;
  FILE *index_fd;

  if ((index_fd = fopen(file, "r")) == NULL) {
    mlog(0, 0, "index_load: No such file '%s'", file);
    return;
  }

  do {
    /*
     * getline() returns a malloc()'d string of 'retlength' characters
     * in 'string_length' sized memory.  We still have to remove the
     * newline though.  The 'retlength' makes that quite painless.
     */
    retlength = getline(&input_line, &string_length, index_fd);
    input_line[retlength - 1] = '\0';

    dlog("input line: %s", input_line);

    if (*input_line == '\0')
      continue;

    /* Support the Diku end-of-file and various line commenting methods. */
    if (*input_line == '$' || *input_line == '#' || (input_line[0] == '/' && input_line[1] == '/'))
      break;

    /* Kick it on down to the single parser. */
    if (!index_load_single(world, input_line))
      mlog(0, 0, "Loading of '%s' failed.", input_line);

  } while (!feof(index_fd));

  /* Don't leak. We didn't use glib to allocate, don't use it to free it. */
  free(input_line);
}

/*
 * Objective: Figure out which file and database method
 * this single zone file wishes to use, and then sic the
 * parser on it.
 */
int index_load_single(World *world, char *line)	/* FIXME: const */
{
  void *handle;
  char *data;
  struct index_file_data zone_info;

  dlog("Handed: '%s'", line);

  /* This will figure out DIO, DFP, room/mobile/object/zone/shop, and the path to feed to DIO. */
  index_file_parse(line, &zone_info);

  /* Ok, now use DIO to grab the data to feed to the parser. */
  if (!(handle = zone_info.dio->open(zone_info.data)))
    return FALSE;

  /* Close DIO on error, we don't access (read: handle) any error information yet. */
  if (!(data = zone_info.dio->data(handle))) {
    zone_info.dio->close(handle);
    return FALSE;
  }

  /* Ok, now feed the information to the appropriate world file parser. */
  zone_info.dfp->parse_by.num[zone_info.type](world, &data);

  /* Clear out the DIO cache and close opened files. */
  zone_info.dio->close(handle);

  return TRUE;
}

/*
 * Objective: Figure out how we will process this line:
 *
 *		Single line description.
 *	<	DBP2 description file.
 *	|	Pipe command generating single line.
 *	|<	Pipe command generating DBP2 description.
 *
 * More possible, those are just ideas.
 */
void index_file_parse(char *line, struct index_file_data *idx_data)
{
  void (*handler)(char *, struct index_file_data *) = NULL;
  void (*pipe_handler)(char *, struct index_file_data *, void (*handler)(char *, struct index_file_data *)) = NULL;

  if (line[0] == '<')
    handler = index_file_parse_file;
  else if (line[0] == '|') {
    pipe_handler = index_file_parse_pipe;
    if (line[1] == '<')
      handler = index_file_parse_file;
    else
      handler = index_file_parse_line;
  } else
    handler = index_file_parse_line;

  /* Skip the type markers and any spaces for the handlers. */
  while (strchr("<| ", *line))
    line++;

  if (pipe_handler)
    (*pipe_handler)(line, idx_data, handler);
  else
    (*handler)(line, idx_data);
}

/*
 * Objective: Intercept program output and feed that to the index routines
 * just as if it was in a file.
 */
void index_file_parse_pipe(char *line, struct index_file_data *idx_data, void (*handler)(char *, struct index_file_data *))
{
  FILE *pipe_str;
  char buffer[256], *endbuf = NULL;
  int bufsize = 1;

  if (!(pipe_str = popen(line, "r"))) {
    perror(line);
    return;
  }

  buffer[255] = '\0';
  do {
    fread(buffer, sizeof(char), 255, pipe_str);
    bufsize += strlen(buffer);
    endbuf = g_realloc(endbuf, bufsize);
    strcat(endbuf, buffer);
    *buffer = '\0';
  } while (!feof(pipe_str));

  pclose(pipe_str);

  /* Woohoo! */
  (*handler)(endbuf, idx_data);

  g_free(endbuf);
}

void index_file_parse_file(char *loadfile, struct index_file_data *idxdata)
{
  struct data_io_operations *fop = dio_lookup("default");
  char *line;
  void *handle;
  int itr;

  struct {
    const char *field;
    int field_offset;
    int type;
  } field_list[] = {
    { "Format", G_STRUCT_OFFSET(struct index_file_data, dfp) , 1 },	// Remove
    { "DataIO", G_STRUCT_OFFSET(struct index_file_data, dio) , 0 },	// magic
    { "Type"  , G_STRUCT_OFFSET(struct index_file_data, type), 2 },	// numbers
    { NULL, -1 }
  };

  if ((handle = fop->open(loadfile)) == NULL) {
    mlog(0, 0, "index_file_parse_file: Failed open of '%s'", line);
    return;
  }

  if ((line = fop->data(handle)) == NULL) {
    mlog(0, 0, "index_file_parse_file: No data!");
    fop->close(handle);
    return;
  }

  memset(idxdata, 0, sizeof(struct index_file_data));

//  strip_comments(&line, CMT_CPP | CMT_SHELL | CMT_C);
  while ((line = sread_line(&line))) {
    dlog("Line: '%s' (%p/%p/%p)", line, idxdata->dio, idxdata->dfp, idxdata->type);

    if (*line == '#' || !strncmp("//", line, 2))
      continue;
    if (*line == '\0')
      break;

    for (itr = 0; field_list[itr].field; itr++) {
      if (g_strncasecmp(line, field_list[itr].field, strlen(field_list[itr].field)))
        continue;

      G_STRUCT_MEMBER(void *, idxdata, field_list[itr].field_offset) = index_line_parse_method(&line, field_list[itr].type);
    }
  }

  if (*line != '\0')
    idxdata->data = g_strdup(line);

  fop->close(handle);
}

/*
 * struct data_io_operations *dio;
 * struct data_parser_operations *dfp;
 * int type;
 * char *data;
 */
void index_file_parse_line(char *line, struct index_file_data *idxdata)
{
  dlog("in: '%s'", line);
  idxdata->dio = index_line_parse_method(&line, ILPM_FILE);
  idxdata->dfp = index_line_parse_method(&line, ILPM_FORMAT);
  idxdata->type = GPOINTER_TO_INT(index_line_parse_method(&line, ILPM_TYPE));
  idxdata->data = line;
  dlog("out: %p/%p/%d '%s'", idxdata->dio, idxdata->dfp, idxdata->type, idxdata->data);
}


void *index_line_parse_method(char **line, int specifier)	/* XXX: enum */
{
  char *ptr, *end;

  typedef void *(*lookup_table_t)(const char *name);

  lookup_table_t type_check_table[] = {
    dio_lookup,
    dfp_lookup,
    dbf_lookup,
    NULL
  };

  dlog("default: %d -> '%p'", specifier, type_check_table[specifier]("default"));

  /* If we don't have a colon, there cannot be a specifier. */ 
  if ((ptr = strchr(*line, ':')) == NULL) {
    if (specifier != ILPM_TYPE)
      return type_check_table[specifier]("default");
    else
      return diku_index_kludge(*line);
  }

  /* Sort of like skip_spaces, but need past the colon. */
  ptr = g_strchug(ptr + 1);

  if ((end = strchr(ptr, ' '))) {
    *end++ = ' ';
    *line = end;
  } else
    *line = ptr + strlen(ptr);

  dlog("found: %d -> '%p'", specifier, type_check_table[specifier](ptr));

  return type_check_table[specifier](ptr);
}

void *diku_index_kludge(const char *filename)
{
  const char *extension[] = { ".mob", ".obj", ".wld", ".shp", ".zon" };
  int i;

  for (i = 0; i < 5; i++)
    if (strstr(filename, extension[i]))
      return GINT_TO_POINTER(i);

  return GINT_TO_POINTER(-1);
}
